//
//  MIDICollection.m
//  MIDIComm
//
//  Created by Matthias Krauß on 30.09.13.
//  Copyright (c) 2013 Matthias Krauß. All rights reserved.
//

#import "MIDICollection.h"
#import <CoreMIDI/CoreMIDI.h>
#include <mach/mach.h>
#include <mach/mach_time.h>

@interface MIDICollection ()

@property (readwrite, strong) NSArray* midiOutEPs;
@property (assign) MIDIClientRef midiClient;
@property (assign) MIDIPortRef midiInPort;
@property (assign) MIDIPortRef midiOutPort;

- (void) midiNotify:(const MIDINotification*)message;
- (void) midiRead:(const MIDIPacketList*)list;

@end

void MyMIDINotifyProc(const MIDINotification* message, void* refCon) {
    MIDICollection* collection = (__bridge MIDICollection*)refCon;
    [collection midiNotify:message];
}

void MyMIDIReadProc(const MIDIPacketList *pktlist, void* readProcRefCon, void* srcConnRefCon) {
    MIDICollection* collection = (__bridge MIDICollection*)srcConnRefCon;
    [collection midiRead:pktlist];
}


@implementation MIDICollection

- (id) init {
    self = [super init];
    if (!self) return nil;
    self.midiOutEPs = @[];

    //Create Client
    MIDIClientRef client;
    OSStatus status = MIDIClientCreate(CFSTR("My Client"), MyMIDINotifyProc, (__bridge void*)self, &client);
    NSAssert(!status, @"MIDIClientCreate failed");
    self.midiClient = client;
    
    //Create Out Port to send
    MIDIPortRef outPort;
    status = MIDIOutputPortCreate (self.midiClient, CFSTR("My Output Port"), &outPort);
    NSAssert(!status, @"MIDIOutputPortCreate failed");
    self.midiOutPort = outPort;

    //Create In Port to receive
    MIDIPortRef inPort;
    status = MIDIInputPortCreate(self.midiClient, CFSTR("My Input Port"), MyMIDIReadProc, (__bridge void*)self, &inPort);
    NSAssert(!status, @"MIDIInputPortCreate failed");
    self.midiInPort = inPort;

    //Iterate through all devices
    NSInteger deviceCount = MIDIGetNumberOfDevices();
    for (int deviceIdx=0; deviceIdx<deviceCount; deviceIdx++) {
        MIDIDeviceRef device = MIDIGetDevice(deviceIdx);
        CFStringRef name = nil;

        //Filter for our device
        if (MIDIObjectGetStringProperty(device, kMIDIPropertyName, &name)) continue;    //Failed to get name
        if (![@"Anykey0x" isEqualToString:(__bridge NSString *)(name)]) continue;       //Not our interface
        
        //Iterate through all entities (logical subcomponents of the device)
        NSInteger entityCount = MIDIDeviceGetNumberOfEntities(device);
        for (int entityIdx=0; entityIdx<entityCount; entityIdx++) {
            MIDIEntityRef entity = MIDIDeviceGetEntity(device, entityIdx);

            //Add all sources to our input port
            NSInteger sourceCount = MIDIEntityGetNumberOfSources(entity);
            for (int srcIdx=0; srcIdx<sourceCount; srcIdx++) {
                MIDIEndpointRef ep = MIDIEntityGetSource(entity, srcIdx);
                NSLog(@"Connecting input to device %i entity %i ep %i",device, entity, ep);
                MIDIPortConnectSource(self.midiInPort, ep, (__bridge void*)self);
            }

            //Remember all destinations to send to them
            NSInteger destinationCount = MIDIEntityGetNumberOfDestinations(entity);
            for (int dstIdx=0; dstIdx<destinationCount; dstIdx++) {
                MIDIEndpointRef ep = MIDIEntityGetDestination(entity, dstIdx);
                NSLog(@"Adding output to device %i entity %i ep %i",device, entity, ep);
                self.midiOutEPs = [self.midiOutEPs arrayByAddingObject:[NSNumber numberWithUnsignedInteger:ep]];
            }
        }
    }
    return self;
}

- (void) dealloc {
    if (self.midiInPort) MIDIPortDispose(self.midiInPort);
    if (self.midiOutPort) MIDIPortDispose(self.midiOutPort);
    if (self.midiClient) MIDIClientDispose(self.midiClient);
}

- (void) sendNoteOnChannel:(int)channel note:(int)note velocity:(int)velocity {
    //Prepare a MIDI packet list with one packet (the note)
    uint8_t buf[4 + 1 * 264];
    MIDIPacketList* mpl = (MIDIPacketList*)buf;
    mpl->numPackets = 1;
    mpl->packet[0].timeStamp = mach_absolute_time();
    mpl->packet[0].length = 3;
    mpl->packet[0].data[0] = 0x90 | (channel & 0x0f);
    mpl->packet[0].data[1] = note & 0xff;
    mpl->packet[0].data[2] = velocity & 0xff;
    
    //Send packet list to all destinations
    for (NSNumber* epNum in self.midiOutEPs) {
        MIDIEndpointRef ep = (MIDIEndpointRef)[epNum unsignedIntegerValue];
        MIDISend(self.midiOutPort, ep, mpl);
    }
}

- (void) sendSysEx:(const uint8_t*)data length:(uint8_t)len {
    //Prepare a MIDI packet list with one packet (our sysex)
    uint8_t buf[4 + 1 * 264];
    if (len > 250) len = 250;
    MIDIPacketList* mpl = (MIDIPacketList*)buf;
    mpl->numPackets = 1;
    mpl->packet[0].timeStamp = mach_absolute_time();
    mpl->packet[0].length = len+5;
    mpl->packet[0].data[0] = 0xf0;  //sysex start
    mpl->packet[0].data[1] = 0x7f;  //universal sysex
    mpl->packet[0].data[2] = 0x7e;  //device ID - something unused, not 7f (sort of broadcast)
    memcpy(&(mpl->packet[0].data[3]), data, len);
    mpl->packet[0].data[3+len] = 0xf7;  //sysex end
    
    //Send packet list to all destinations
    for (NSNumber* epNum in self.midiOutEPs) {
        MIDIEndpointRef ep = (MIDIEndpointRef)[epNum unsignedIntegerValue];
        MIDISend(self.midiOutPort, ep, mpl);
    }
}

//Called when a midi message comes in. Filter packets for our SysEx response
- (void) midiRead:(const MIDIPacketList*)list {
    @autoreleasepool {
        for (int i=0; i<list->numPackets; i++) {
            int pktLen = list->packet[i].length;
            if (pktLen < 4) continue;
            const uint8_t* buf = (list->packet[i].data);
            if ((buf[0] == 0xf0) && (buf[1] == 0x7f) && (buf[2] == 0x7e) && (buf[pktLen-1] == 0xf7)) {
                NSString* s = [[NSString alloc] initWithBytes:&(buf[3]) length:pktLen-4 encoding:NSASCIIStringEncoding];
                [self gotReply:s];
            }
        }
    }
}

- (void) gotReply:(NSString*)reply {
    if ([self.delegate respondsToSelector:@selector(gotReply:)]) [self.delegate gotReply:reply];
}

- (void) midiNotify:(const MIDINotification*)message {
    NSLog(@"MIDI notification %i",message->messageID);
}



@end
